/**
 * \file tab.cpp
 * Tab contains all widgets within a single tab.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 8 Dec 2009
 *
 * Copyright (C) 2009  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "tab.h"
#include "tablewidget.h"
#include "application.h"
#include "mainwindow.h"
#include "filesrenamer.h"
#include "tagsfiller.h"
#include "tagseditor.h"
#include "lineedit.h"
#include "undoactions.h"
#include "qoobarglobals.h"
#include "imagebox.h"
#include "treewidget.h"
#include "tablewidget.h"
#include "newtagdialog.h"
#include "macsplitter.h"
#include "replaygaindialog.h"
//#include "delayedexecutiontimer.h"
#include "mp3tagsdialog.h"
#include "iqoobarplugin.h"

#ifdef HAVE_QT5
#include <QtWidgets>
#include <QtConcurrent/QtConcurrent>
#else
#include <QtGui>
#endif
#include "enums.h"
#include "stringroutines.h"

#include "model.h"

QList<BufferItem> Tab::internalBuffer = QList<BufferItem>();

class ChangeFilesHelper {
public:
    ChangeFilesHelper(Model *model, const QList<int> &rows)
    {
        _indexes = model->selection();
        oldTags = model->selectedFiles();
        newTags=oldTags;
        _rows=rows;
    }

    void add(const int row, const QString &key, const QStringList &newValues)
    {
        int size=qMin(_indexes.size(), newValues.size());
        for (int i=0; i<size; ++i) {
            if (row<TAGSCOUNT) newTags[i].setTag(row, newValues.at(i));
            else newTags[i].setUserTag(key, newValues.at(i));
        }
    }

    QUndoCommand *getCommand(const QString &title,Tab *t)
    {
        if (oldTags==newTags) return 0;
        return new ChangeFilesCommand(_indexes,oldTags,newTags,_rows,title,t);
    }

private:
    QList<Tag> oldTags;
    QList<Tag> newTags;
    QList<int> _indexes;
    QList<int> _rows;
};

void Tab::changeRow(ChangeFilesHelper *helper, int row, const QString &s, const QStringList &newValues)
{
    if (App->useUndo && helper)
        helper->add(row,s,newValues);
    else {
        QList<int> indexes = model->selection();
        int size=qMin(indexes.size(), newValues.size());
        Q_ASSERT(tree->topLevelItemCount()>=size);
        for (int i=0; i<size; ++i) {
            if (row<TAGSCOUNT) {
                model->setTag(indexes.at(i), row, newValues.at(i));
                tree->topLevelItem(indexes.at(i))->setText(row,newValues.at(i).simplified());
            }
            else
                model->setTag(indexes.at(i), s, newValues.at(i));
        }
        table->blockSignals(true);
        updateTagsTableRow(row, newValues);
        table->blockSignals(false);
    }
}

void Tab::pushCommand(ChangeFilesHelper *helper, const QString &motivation)
{
    if (App->useUndo) {
        QUndoCommand *c=helper->getCommand(motivation,this);
        if (c) undoStack_->push(c);
        delete helper;
    }
    else {
        Q_FOREACH (const int &i, model->selection()) {
            if (model->fileAt(i).wasChanged())
                tree->updateFileSaveStatus(i,true);
            qApp->processEvents();
        }

        emit filesChanged(!allFilesSaved());
    }
}

//void Tab::setTableUpdateMode(bool useDelayedUpdate)
//{
//    tree->disconnect(SIGNAL(itemSelectionChanged()));
//    if (useDelayedUpdate) {
//        if (!delay) delay = new DelayedExecutionTimer(500,150,this);
//        connect(tree,SIGNAL(itemSelectionChanged()), delay,SLOT(trigger()));
//        connect(delay,SIGNAL(triggered()),SLOT(filesSelectionChanged()));
//    }
//    else {
//        connect(tree,SIGNAL(itemSelectionChanged()), SLOT(filesSelectionChanged()));
//    }
//}

Tab::Tab(MainWindow *parent) : QWidget(parent)
{
    undoStack_ = new QUndoStack(this);


    wasRenamed = false;

    win=parent;
    //    delay=0;
    //    setTableUpdateMode(App->useDelayedUpdate);
    tree = new TreeWidget(this);
    connect(tree,SIGNAL(itemSelectionChanged()), SLOT(filesSelectionChanged()));
    connect(tree,SIGNAL(itemEntered(QTreeWidgetItem *,int)),SLOT(updateStatusBar(QTreeWidgetItem *)));
    connect(tree,SIGNAL(itemClicked(QTreeWidgetItem *,int)),SLOT(onItemClicked(QTreeWidgetItem *,int)));
    connect(tree,SIGNAL(moveToTab(int)),this,SLOT(moveToTab(int)));
    connect(tree,SIGNAL(showMP3TagsDialog()),SLOT(showMP3TagsDialog()));
    //**************************************************************************
    table = new TableWidget(this);
    connect(table,SIGNAL(itemChanged(QTableWidgetItem *)),SLOT(cellChanged(QTableWidgetItem *)));
    connect(table,SIGNAL(cellClicked()),SLOT(editCell()));
    connect(table,SIGNAL(itemSelectionChanged()),SLOT(handleCutCopy()));
    //=====================================================
    imageBox=new ImageBox(this);
    connect(imageBox,SIGNAL(imageChanged(CoverImage,QString)),SLOT(changeImage(CoverImage,QString)));

    showImageBoxButton = new QPushButton(qApp->style()->standardIcon(QStyle::SP_ArrowLeft),QString(),this);
    showImageBoxButton->setFlat(true);
    showImageBoxButton->setFixedSize(20,20);
    showImageBoxButton->setFocusPolicy(Qt::NoFocus);
    showImageBoxButton->setToolTip(tr("Cover Art"));
    connect(showImageBoxButton,SIGNAL(clicked()),SLOT(toggleImageBoxVisibility()));

    //=======================================================
    model = new Model(this);
    connect(model,SIGNAL(fileChanged(int,bool)),tree,SLOT(updateFileSaveStatus(int,bool)), Qt::QueuedConnection);
    connect(model,SIGNAL(fileChanged(int,Tag,QList<int>)),tree,SLOT(updateRow(int,Tag,QList<int>)));
    connect(model,SIGNAL(imageChanged(int,bool)),tree,SLOT(updateImage(int,bool)));
    connect(model,SIGNAL(fileNameChanged(int,QString)),tree,SLOT(updateFileName(int,QString)));
    connect(model,SIGNAL(message(int,QString)),SLOT(showMessage(int,QString)));

    QHBoxLayout *bL=new QHBoxLayout;
    bL->setContentsMargins(0,0,0,0);
    bL->addWidget(table);
    bL->addWidget(showImageBoxButton);
    bL->addWidget(imageBox);
    QWidget *wi=new QWidget;
    wi->setLayout(bL);

#ifndef Q_OS_MAC
    QSplitter *sp = new QSplitter(Qt::Vertical,this);
#else
    QSplitter *sp = new MacSplitter();
    sp->setOrientation(Qt::Vertical);
#endif
    sp->setOpaqueResize(false);
    sp->addWidget(tree);
    sp->addWidget(wi);
#ifndef Q_OS_MAC
    sp->setContentsMargins(0,0,0,0);
#endif
    if (!App->splitterState.isEmpty())
        sp->restoreState(App->splitterState);
    connect(sp,SIGNAL(splitterMoved(int,int)),SLOT(saveSplitterState()));

    QVBoxLayout *la=new QVBoxLayout;
    la->setContentsMargins(0,0,0,0);
    la->addWidget(sp);
    setLayout(la);

    imageBox->hide();

    moveUpAct = new QAction(qApp->style()->standardIcon(QStyle::SP_ArrowUp),tr("Move up"),this);
    connect(moveUpAct,SIGNAL(triggered()),this,SLOT(moveUp()));
    moveUpAct->setShortcut(tr("Ctrl+Up"));
    moveUpAct->setDisabled(true);

    moveDownAct = new QAction(qApp->style()->standardIcon(QStyle::SP_ArrowDown),tr("Move down"),this);
    connect(moveDownAct,SIGNAL(triggered()),this,SLOT(moveDown()));
    moveDownAct->setShortcut(tr("Ctrl+Down"));
    moveDownAct->setDisabled(true);

    editAct = new QAction(tr("&Edit..."),this);
    connect(editAct,SIGNAL(triggered()),this,SLOT(editCell()));

    multilineEditAct  = new QAction(tr("Edit as &multiline text..."),this);
    connect(multilineEditAct,SIGNAL(triggered()),this,SLOT(editCell()));


    removeAct = new QAction(tr("C&lear tags"),this);
#if QT_VERSION >= 0x040600
    removeAct->setIcon(QIcon::fromTheme(QSL("edit-delete"),QIcon(QSL(":/src/icons/remove_tags.png"))));
#else
    removeAct->setIcon(QIcon(QSL(":/src/icons/remove_tags.png")));
#endif
    removeAct->setShortcut(Qt::Key_Delete);
    removeAct->setShortcutContext(Qt::WidgetShortcut);
    connect(removeAct,SIGNAL(triggered()),this,SLOT(remove()));

    collectAct = new QAction(tr("Collect for &future use"),this);
    collectAct->setShortcut(tr("Ctrl+H"));
    connect(collectAct,SIGNAL(triggered()),this,SLOT(collectTags()));

    tree->addAction(win->actions[QSL("rereadTags")]);
    tree->addAction(win->actions[QSL("rename")]);
    tree->addAction(win->actions[QSL("fill")]);
    tree->addAction(win->actions[QSL("delFiles")]);
    tree->addAction(win->actions[QSL("delAllFiles")]);
    tree->addAction(moveUpAct);
    tree->addAction(moveDownAct);
    tree->addAction(win->actions[QSL("play")]);
    tree->addAction(win->actions[QSL("replaygain")]);

    table->addAction(editAct);
    table->addAction(multilineEditAct);
    table->addAction(win->actions[QSL("cut")]);
    table->addAction(win->actions[QSL("copy")]);
    table->addAction(win->actions[QSL("paste")]);
    table->addAction(win->actions[QSL("copyToClipboard")]);
    table->addAction(win->actions[QSL("pasteFromClipboard")]);
    table->addAction(removeAct);
    table->addAction(win->actions[QSL("removeTags")]);
    table->addAction(win->actions[QSL("newTag")]);
    table->addAction(collectAct);

    operationsMapper = new QSignalMapper(this);
    connect(operationsMapper,SIGNAL(mapped(QString)),SLOT(handleOperation(QString)));
    operationsAct = new QAction(tr("Operations"),this);
    table->addAction(operationsAct);

    emit totalLengthChanged(0);
    tree->setFocus();
}

void Tab::retranslateUi()
{
    showImageBoxButton->setToolTip(tr("Cover Art"));
    moveUpAct->setText(tr("Move up"));
    moveUpAct->setShortcut(tr("Ctrl+Up"));
    moveDownAct->setText(tr("Move down"));
    moveDownAct->setShortcut(tr("Ctrl+Down"));
    editAct->setText(tr("&Edit..."));
    multilineEditAct->setText(tr("Edit as &multiline text..."));

    removeAct->setText(tr("C&lear tags"));
    collectAct->setText(tr("Collect for &future use"));
    collectAct->setShortcut(tr("Ctrl+H"));

    operationsAct->setText(tr("Operations"));
    QMenu *menu=operationsAct->menu();
    delete menu;
    operationsAct->setMenu(createOperations(operationsMapper,this));

    tree->retranslateUi();
    table->retranslateUi();
    imageBox->retranslateUi();
}



void Tab::saveSplitterState() /*SLOT*/
{
    QSplitter *sp=qobject_cast<QSplitter *>(sender());
    if (sp)
        App->splitterState=sp->saveState();
}

bool Tab::modelNotEmpty()
{
    return !model->isEmpty();
}


/**
  Editing functions
  */

void Tab::handleOperation(const QString &type) /*SLOT*/
{
    if (!model->hasSelection()) return;

    QList<int> rows=table->selectedRows();
    if (rows.isEmpty()) return;

    ChangeFilesHelper *helper=0;
    if (App->useUndo)
        helper = new ChangeFilesHelper(model,rows);

    Q_FOREACH (const int &row,rows) {
        QStringList newValues=getTags(row); //tags to script
        newValues=QtConcurrent::blockingMapped(newValues,Operated(type));
        changeRow(helper,row,table->tag(row),newValues);
    }
    pushCommand(helper,tr("changing tags"));
}

void Tab::tagsChanged(int row,const QStringList &newValues) /*SLOT*/
{
    if (!model->hasSelection()) return;

    ChangeFilesHelper *helper=0;
    if (App->useUndo)
        helper = new ChangeFilesHelper(model, QList<int>()<< row);
    changeRow(helper,row,table->tag(row),newValues);
    pushCommand(helper,tr("changing tags"));

    if (App->collectSilently && row<TAGSCOUNT)
        if (App->useCompletion.at(row))
            App->updateCompletions(row, newValues);
}




void Tab::setTags(const QList<int> &inds, const QList<Tag> &ts, const QList<int> &rows)
{
    if (inds.isEmpty()) return;

    model->setTags(inds, ts);
    updateFilesTable(inds, rows);
    if (model->hasSelection())
        updateTagsTable(rows);
}

void Tab::updateFilesTable(const QList<int> &inds, const QList<int> &rows)
{
    Q_FOREACH (const int &i, inds)
        tree->updateRow(i, model->fileAt(i), rows);
    emit filesChanged(!allFilesSaved());
}

struct Reduce
{
    Reduce(int size)
        : m_size(size) { }

    typedef QString result_type;

    QString operator()(const QStringList &list)
    {
        if (list.isEmpty()) return QString();
        QStringList l;
        if (list.size()<=50) {
            l=list;
            l.removeDuplicates();
        }
        else {
            //reduce to a string with length not greater than table width
            Q_FOREACH (const QString &s, list) {
                if (!needToAdd(l)) break;
                if (!l.contains(s)) l<<s;
            }
        }
        if (l.size()==1) return l.first();
        return QObject::tr("<<multiple>>") + l.join(QSL(";"));
    }

    bool needToAdd(const QStringList &list) const {
        return qApp->fontMetrics().width(list.join(QSL(";"))) < m_size;
    }

    int m_size;
};

void Tab::updateTagsTable(const QList<int> &rows)
{
    bool fullUpdate = (rows.isEmpty() || rows.last()>=TAGSCOUNT);

    int indexesCount = model->selectionSize();
    table->blockSignals(true);
    table->setCount(indexesCount);
    if (fullUpdate) {
        table->setRowCount(TAGSCOUNT);
        for (int i=0; i<TAGSCOUNT; ++i)
            table->item(i,0)->setText(QString());
    }

    if (indexesCount>0) {
        QList<int> _rows;
        if (fullUpdate) {
            QStringList otherid = model->userTagsIds();
            Q_FOREACH(const QString &s,otherid)
                table->addRow(s);
            for (int i=0; i<table->rowCount(); ++i) _rows << i;
        }
        else _rows=rows;

        QList<QStringList> tableTags;
        Q_FOREACH (const int &i,_rows)
            tableTags << getTags(i);

        QStringList reducedTags = QtConcurrent::blockingMapped(tableTags,Reduce(table->width()));

        table->blockSignals(true);
        for (int i=0; i<_rows.size(); ++i) {
            table->updateRow(_rows.at(i),reducedTags.at(i));
        }
        table->blockSignals(false);
    }

    table->blockSignals(false);
    table->resizeColumnsToContents();
    updateImageBox();
}

void Tab::updateTagsTableRow(int row, const QStringList &list)
{
    table->blockSignals(true);
    QString text=Reduce(table->width())(list);
    table->updateRow(row, text);
    table->blockSignals(false);
}

void Tab::remove() /*SLOT*/
{
    if (!model->hasSelection()) return;
    QList<int> rows=table->selectedRows();
    if (rows.isEmpty()) return;

    QStringList newValues=QVector<QString>(model->selectionSize()).toList();

    ChangeFilesHelper *helper=0;
    if (App->useUndo)
        helper = new ChangeFilesHelper(model,rows);

    Q_FOREACH (const int &row,rows)
        changeRow(helper,row,table->tag(row),newValues);

    pushCommand(helper,tr("clearing tags"));
}

void Tab::removeAllTags() /*SLOT*/
{
    table->selectAll();
    remove();
}

void Tab::editCell() /*SLOT*/
{
    if (!model->hasSelection()) return;
    int row = table->currentRow();

    bool plain=true;
    QAction *a=qobject_cast<QAction *>(sender());
    if (a) plain = (a!=multilineEditAct);

    QStringList oldValues = getTags(row);

    TagsEditDialog editor(row,table->verticalHeaderItem(row)->text(),
                          oldValues,table->item(row,0)->text(),plain,this);
    editor.setModel(model);
    connect(&editor,SIGNAL(rowSelected(int)),tree,SLOT(synchronizeLine(int)));
    connect(&editor,SIGNAL(destroyed()),tree,SLOT(editingDone()));
    connect(&editor,SIGNAL(tagsSent(int,QStringList)),SLOT(tagsChanged(int,QStringList)));

    if (editor.exec()) {
        QStringList newValues=editor.getList();
        if (newValues==oldValues) return;

        tagsChanged(row,newValues);
    }
}

void Tab::updateTags(QList<Tag> &oldTags, QList<Tag> &newTags, const QString &motivation)
{
    QList<int> inds;
    QList<int> indexes=model->selection();
    for (int i=oldTags.size()-1; i>=0; --i) {
        if (oldTags.at(i)==newTags.at(i)) {
            oldTags.removeAt(i);
            newTags.removeAt(i);
        }
        else
            inds.prepend(indexes.at(i));
    }

    if (inds.isEmpty()) return;
    if (App->useUndo) {
        ChangeFilesCommand *command = new ChangeFilesCommand(inds,
                                                             oldTags,
                                                             newTags,
                                                             QList<int>(),
                                                             motivation,
                                                             this);
        undoStack_->push(command);
    }
    else {
        model->setTags(inds,newTags);
        updateTagsTable(QList<int>());
    }
}

void Tab::fill() /*SLOT*/
{
    if (!model->hasSelection())
        return;

    QList<Tag> oldTags = model->selectedFiles();

    TagsFillDialog filler(oldTags,this);
    if (filler.exec()) {
        QList<Tag> newTags=filler.getNewTags();
        updateTags(oldTags, newTags, tr("filling tags"));
    }
}

void Tab::doPlugin(IQoobarPlugin *plugin)
{
    if (!model->hasSelection())  return;

    QList<Tag> oldTags=model->selectedFiles();
    QList<Tag> newTags = plugin->getNewTags(oldTags);
    updateTags(oldTags, newTags, tr("changing tags"));
}

void Tab::rereadTags() /*SLOT*/
{
    if (!model->hasSelection()) return;

    if (App->useUndo) {
        QList<Tag> oldTags = model->selectedFiles();
        QStringList fileNames = model->fileNames();
        QList<Tag> newTags = QtConcurrent::blockingMapped(fileNames, readTags);

        QList<int> inds;
        QList<int> indexes=model->selection();
        for (int i=oldTags.size()-1; i>=0; --i) {
            if (oldTags.at(i)==newTags.at(i)) {
                oldTags.removeAt(i);
                newTags.removeAt(i);
            }
            else
                inds.prepend(indexes.at(i));
        }
        if (inds.isEmpty()) return;

        ChangeFilesCommand *command = new ChangeFilesCommand(inds,
                                                             oldTags,
                                                             newTags,
                                                             QList<int>(),
                                                             tr("rereading tags"),
                                                             this);
        undoStack_->push(command);
        emit filesChanged(!allFilesSaved());
    }
    else
        rereadTags(false, false);
}

void Tab::rereadTags(bool all, bool force)
{
    model->rereadTags(all, force);
    emit filesChanged(!allFilesSaved());
    updateTagsTable(QList<int>());
}


void Tab::cellChanged(QTableWidgetItem *item) /*SLOT*/
{
    if (!model->hasSelection()) return;
    if (!item) return;
    int row=item->row();

    QStringList oldValues = getTags(row);
    QStringList newValues=model->tagsByPattern(row, item->text());
    table->closePersistentEditor(item);
    if (oldValues==newValues) {
        updateTagsTableRow(row, newValues);
    }
    else {
        tagsChanged(row, newValues);
    }
}

void Tab::paste() /*SLOT*/
{
    if (!model->hasSelection()) return;

    ChangeFilesHelper *helper=0;
    if (App->useUndo)
        helper = new ChangeFilesHelper(model,QList<int>());

    Q_FOREACH (const BufferItem &item, internalBuffer) {
        int row=-1;
        QString s=item.header;
        if (internalBuffer.size()==1) row = table->currentRow();
        else {
            row=item.row;
            if (row>=TAGSCOUNT) {
                int found=-1;
                int c=table->rowCount();
                for (int j=0; j<c; ++j)
                    if (table->tag(j) == s) {
                        found=j;
                        break;
                    }
                if (found<0) row=c;
                else row=found;
            }
        }
        if (row==-1 || row>=table->rowCount()) continue;

        QStringList newValues=item.values;
        if (newValues.size()==1 && App->autoexpand)
            newValues = QVector<QString>(model->selection().size(),newValues.first()).toList();

        changeRow(helper,row,s,newValues);
    }
    pushCommand(helper,tr("pasting tags"));
}

void Tab::pasteFromClipboard() /*SLOT*/
{

    if (!model->hasSelection()) return;
    int row = table->currentRow();
    QList<int> rows; rows << row;

    QList<int> indexes = model->selection();
    QClipboard *clipboard = QApplication::clipboard();
    QString text = clipboard->text();
    if (text.isEmpty()) return;
    QStringList newValues=text.split(QRegExp(QSL("\\n|\\r|\\n\\r|\\r\\n")));
    if (newValues.size()==1 && App->autoexpand) {
        while (indexes.size()>newValues.size()) newValues<<newValues.first();
    }
    while (indexes.size()<newValues.size()) newValues.removeLast();
    while (indexes.size()>newValues.size()) newValues<<QString();


    ChangeFilesHelper *helper=0;
    if (App->useUndo)
        helper = new ChangeFilesHelper(model,rows);

    Q_FOREACH (const int &row,rows) {
        changeRow(helper,row,table->tag(row),newValues);
    }
    pushCommand(helper,tr("pasting tags"));
}

void Tab::cut() /*SLOT*/
{
    copy();
    remove();
}

void Tab::copy() /*SLOT*/
{
    if (!model->hasSelection()) return;
    internalBuffer.clear();
    for (int row=0; row<table->rowCount(); ++row) {
        if (table->item(row,0)->isSelected()) {
            internalBuffer.append(BufferItem(row,table->tag(row), getTags(row)));
        }
    }
    emit bufferChanged(internalBuffer.isEmpty());
}

void Tab::copyToClipboard() /*SLOT*/
{
    if (!model->hasSelection()) return;
    QList<int> indexes = model->selection();

    QClipboard *clipboard = QApplication::clipboard();
    QStringList result;
    Q_FOREACH (const int &index, indexes) {
        QStringList s;
        for (int row=0; row<table->rowCount(); ++row) {
            if (table->item(row,0)->isSelected()) {
                if (row<TAGSCOUNT)
                    s<<model->value(index,row);
                else s << model->value(index,table->tag(row));
            }
        }
        result << s.join(QSL("\t"));
    }
    clipboard->setText(result.join(QSL("\n")));
}


















/**
  Images
  */

void Tab::changeImage(const CoverImage &img,const QString &message) /*SLOT*/
{
    if (!model->hasSelection()) return;

    if (App->useUndo) {
        ChangeCommand *command = new ChangeCommand(this);
        command->setText(message);

        QList<int> indexes = model->selection();
        int added=0;
        Q_FOREACH(const int &i, indexes) {
            CoverImage oldImg = model->fileAt(i).image();
            if (img==oldImg) continue;
            added++;
            new ChangeImageCommand(i,oldImg,img,this,command);
        }
        if (added>0)
            undoStack_->push(command);
        else delete command;
    }
    else {
        model->setImage(img);
        setStatus();
    }
}

void Tab::setStatus()
{
    emit filesChanged(!allFilesSaved());
    updateImageBox();
}

void Tab::toggleImageBoxVisibility() /*SLOT*/
{
    if (imageBox->isHidden()) {
        imageBox->show();
        showImageBoxButton->setIcon(qApp->style()->standardIcon(QStyle::SP_ArrowRight));
        updateImageBox();
    }
    else {
        imageBox->hide();
        showImageBoxButton->setIcon(qApp->style()->standardIcon(QStyle::SP_ArrowLeft));
    }
}

void Tab::updateImageBox()
{
    if (!imageBox->isVisible()) return; //do not paint image if it is not visible
    imageBox->clear();

    if (!model->hasSelection()) return;

    if (!model->imagesAreSame()) //images are different
        imageBox->updatePicture(tr("<<multiple>>"));
    else {//images are the same
        CoverImage img=model->fileAtSelection(0).image();
        imageBox->updatePicture(img,model->fileAtSelection(0).fullFileName());
    }
}




/**
  ***********************************************
  */

void Tab::play() /*SLOT*/
{
    QString s=App->player;
    if (s.isEmpty()) {
        QMessageBox::warning(this,QSL("Qoobar"),tr("Please set a player in the Settings dialog"));
        return;
    }
    QStringList list=model->fileNames();
    if (!list.isEmpty())
        QProcess::startDetached(s,list);
}


/**
  Rename files
  */

void Tab::renameFiles() /*SLOT*/
{
    if (!model->hasSelection()) return;

    QStringList oldFileNames = model->fileNames();

    FileRenameDialog renamer(model,oldFileNames,this);
    if (renamer.exec()) {
        QStringList newFileNames = renamer.getNewFileNames();
        if (newFileNames == oldFileNames) return;

        QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
        model->rename(newFileNames);
        QApplication::restoreOverrideCursor();
        undoStack_->clear();
    }
}



/**
  Add and del files
  */
void Tab::delFiles() /*SLOT*/
{
    delFiles(false);
}

void Tab::delAllFiles() /*SLOT*/
{
    model->selectAll();
    delFiles(false);
}

void Tab::delFiles(bool deleteSilently)
{
    QList<int> inds = model->selection();
    if (model->hasSelection()) {
        if (!deleteSilently && !model->selectionSaved()) {
            QMessageBox msgBox(QMessageBox::Question,tr("Qoobar"),
                               tr("Some files have been modified"),
                               QMessageBox::NoButton);
            msgBox.setInformativeText(tr("Remove them anyway?"));
            QPushButton *delB=msgBox.addButton(tr("Yes, remove without saving"),QMessageBox::YesRole);
            delB->setIcon(qApp->style()->standardIcon(QStyle::SP_DialogSaveButton));
            QPushButton *saveB=msgBox.addButton(tr("No, save them and then remove"),QMessageBox::NoRole);
            saveB->setIcon(qApp->style()->standardIcon(QStyle::SP_DialogDiscardButton));
            QPushButton *cancB=msgBox.addButton(QMessageBox::Cancel);
            cancB->setIcon(qApp->style()->standardIcon(QStyle::SP_DialogCancelButton));
            msgBox.setDefaultButton(saveB);
            msgBox.setEscapeButton(cancB);
            msgBox.exec();

            if (msgBox.clickedButton() == saveB) {
                model->save();
            }
            else if (msgBox.clickedButton() == cancB) {
                return;
            }
        }
        model->delFiles();
        if (!inds.isEmpty()) {
            tree->blockSignals(true);

            int count=inds.size()-1;
            for (int i=count; i>0; --i) {
                delete tree->takeTopLevelItem(inds.at(i));
            }
            tree->blockSignals(false);
            delete tree->takeTopLevelItem(inds.at(0));
            undoStack_->clear();
        }
        emit filesChanged(!allFilesSaved());
        emit filesCountChanged(modelNotEmpty());
        emit updateStatusBar(Tag());
        updateTagsTable(QList<int>());
    }
}

void Tab::addFiles(const QStringList &filesToAdd, bool clearBefore)
{
    if (filesToAdd.isEmpty()) return;
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    if (clearBefore) delAllFiles();

    QList<Tag> tagsToAdd = model->addFiles(filesToAdd);
    tree->addRows(&tagsToAdd);
    emit filesCountChanged(true);
    //emit filesChanged(!allFilesSaved());
    undoStack_->clear();
    QApplication::restoreOverrideCursor();
}

void Tab::addFiles(const QList<Tag> &tagsToAdd)
{
    model->addFiles(tagsToAdd);
    tree->addRows(&tagsToAdd);
    undoStack_->clear();
    emit filesCountChanged(true);
    emit filesChanged(!allFilesSaved());
}




/**
Saving files
*/
void Tab::saveTags() /*SLOT*/
{
    if (model->allSaved()) return;
    model->save();
    rereadTags(true);
    undoStack_->clear();
}

void Tab::onItemClicked(QTreeWidgetItem *item,int col) /*SLOT*/
{
    if (!item) return;
    if (col==TAGSCOUNT+2) {
        if (model->saveAt(tree->indexOfTopLevelItem(item))) {
            rereadTags(false);
            undoStack_->clear();
        }
    }
}

bool Tab::maybeSave(const QString &tabText)
{

    if (App->saveChanges) {
        saveTags();
        return true;
    }

    if (!allFilesSaved()) {
        QMessageBox msgBox(QMessageBox::Question,
                           tr("Closing Qoobar"),
                           tr("Files in %1 have been modified.").arg(tabText),
                           QMessageBox::NoButton,
                           this);
        msgBox.setInformativeText(tr("Do you want to save your changes?"));
        QPushButton *saveB=msgBox.addButton(tr("Yes, save"),QMessageBox::YesRole);
        saveB->setIcon(qApp->style()->standardIcon(QStyle::SP_DialogSaveButton));
        QPushButton *discB=msgBox.addButton(tr("No, don't save"),QMessageBox::NoRole);
        discB->setIcon(qApp->style()->standardIcon(QStyle::SP_DialogDiscardButton));
        QPushButton *cancB=msgBox.addButton(QMessageBox::Cancel);
        cancB->setIcon(qApp->style()->standardIcon(QStyle::SP_DialogCancelButton));
        msgBox.setDefaultButton(saveB);
        msgBox.setEscapeButton(cancB);
        msgBox.exec();

        if (msgBox.clickedButton() == cancB) return false;
        else {
            if (msgBox.clickedButton() == saveB) saveTags();
            return true;
        }
    }
    return true;
}

bool Tab::allFilesSaved() const
{
    return model->allSaved();
}

bool Tab::filesSelected()
{
    return model->hasSelection();
}

/**
  Moving files up and down
  */

void Tab::moveUp() /*SLOT*/
{
    moveItems(true);
}

void Tab::moveDown() /*SLOT*/
{
    moveItems(false);
}

void Tab::moveItems(bool up)
{
    QList<int> indexes = model->selection();
    model->move(up);
    QList<int> newIndexes = model->selection();

    tree->blockSignals(true);

    int i=up?0:indexes.size()-1;
    while (1) {
        tree->insertTopLevelItem(newIndexes.at(i),tree->takeTopLevelItem(indexes.at(i)));
        if ((up && i==indexes.size()-1) || (!up && i==0)) break;
        i=up?i+1:i-1;
    }

    tree->clearSelection();
    Q_FOREACH (const int &i, newIndexes)
        tree->topLevelItem(i)->setSelected(true);
    tree->blockSignals(false);
    undoStack_->clear();
}

void Tab::moveToTab(int tabID) /*SLOT*/
{
    if (!model->hasSelection()) return;
    QList<Tag> filesToMove = model->selectedFiles();

    emit moveFilesToTab(tabID, filesToMove);
    delFiles(true);
}


/**
  Other
  */

void Tab::filesSelectionChanged() /*SLOT*/
{
    QList<int> indexes;

    for (int i=0; i<tree->topLevelItemCount(); ++i) {
        if (tree->topLevelItem(i)->isSelected()) {
            indexes << i;
        }
    }
    model->setSelection(indexes);

    emit selectionChanged(indexes.size()>0);
    emit totalLengthChanged(model->totalLength());

    moveUpAct->setEnabled(indexes.size()>0);
    moveDownAct->setEnabled(indexes.size()>0);

    updateTagsTable(QList<int>());
    handleCutCopy();
}

QStringList Tab::getTags(int n)
{
    if (n<TAGSCOUNT)
        return model->values(n);
    else
        return model->values(table->tag(n));
}

void Tab::newTag() /*SLOT*/
{
    if (!model->hasSelection()) return;

    QString id;
    NewTagDialog dialog(this);
    if (dialog.exec())
        id = dialog.newTagName.toUpper();
    if (id.isEmpty()) return;

    int found=-1;
    int c=table->rowCount();
    for (int i=0; i<c; ++i)
        if (table->verticalHeaderItem(i)->text()==id) {
            found=i;
            break;
        }
    if (found<0) {
        table->addRow(id);
        found=c;
    }

    table->setCurrentItem(table->item(found,0),QItemSelectionModel::ClearAndSelect);
    table->resizeColumnsToContents();
}

void Tab::handleCutCopy() /*SLOT*/
{
    bool b=(!table->selectedItems().isEmpty() && model->hasSelection());

    emit tagsSelectionChanged(b);

    editAct->setEnabled(b);
    multilineEditAct->setEnabled(b);
    removeAct->setEnabled(b);
    operationsAct->setEnabled(b);
    collectAct->setEnabled(b);
}

void Tab::replaygain() /*SLOT*/
{
    if (!model->hasSelection()) return;

    //we can only scan saved files, because we need to write
    //replaygain info
    if (!model->selectionSaved()) {
        if (QMessageBox::question(this,tr("Qoobar - ReplayGain Info"),
                                  tr("Qoobar can only scan saved files.\n"
                                     "Do you want to save changes?"),
                                  QMessageBox::Yes | QMessageBox::No,
                                  QMessageBox::Yes) == QMessageBox::Yes) {
            model->save();
            rereadTags(false);
        }
        else return;
    }

    ReplayGainDialog dialog(model,this);
    dialog.exec();
    //updateFilesTable(model->selection(), QList<int>());
    Q_FOREACH (const int &i, model->selection())
        tree->updateRG(i, model->fileAt(i));
    emit filesChanged(!allFilesSaved());
}

void Tab::updateStatusBar(QTreeWidgetItem *item) /*SLOT*/
{
    int row=tree->indexOfTopLevelItem(item);
    if (row<0)
        return;

    emit updateStatusBar(model->fileAt(row));
}

void Tab::collectTags() /*SLOT*/
{
    if (!model->hasSelection()) return;

    for (int i=0; i<TAGSCOUNT; ++i) {
        if (table->item(i,0)->isSelected()) {
            if (table->item(i,0)->text().isEmpty()) continue;
            App->updateCompletions(i,getTags(i));
        }
    }
}

QMap<int, QString> Tab::allTabsNames()
{
    return win->allTabsNames();
}

bool Tab::allSelectedFilesAreMP3() const
{
    return model->allSelectedFilesAreMP3();
}

QUndoStack *Tab::undoStack()
{
    return undoStack_;
}

void Tab::showMP3TagsDialog() /*SLOT*/
{
    if (!model->hasSelection()) return;

    MP3TagsDialog dialog(model->selectionSize(),this);

    if (dialog.exec()) {
        //changing global settings just for this only operation
        int tagTypes = dialog.tagTypes();
        int id3v1Synchro = App->id3v1Synchro;
        App->id3v1Synchro = (tagTypes & MP3TagsDialog::ID3V1)
                ?ID3V1_UPDATE_ALWAYS
               :ID3V1_DELETE;
        bool mp3writeape = App->mp3writeape;
        App->mp3writeape = tagTypes & MP3TagsDialog::APE;
        bool mp3writeid3 = App->mp3writeid3;
        App->mp3writeid3 = tagTypes & MP3TagsDialog::ID3V2;

        //now we are ready to save files
        //we just have to ensure files will be saved
        model->forceSelectionToSave();
        model->save();

        undoStack_->clear();
        rereadTags(false);


        //returning back global settings
        App->id3v1Synchro= id3v1Synchro;
        App->mp3writeape = mp3writeape;
        App->mp3writeid3 = mp3writeid3;
    }
}

void Tab::showMessage(int type, const QString &text)
{
    switch (type) {
    case MT_ERROR: QMessageBox::critical(this, tr("Qoobar"), text);
        break;
        //do not show successful result, as we are not in console mode
//    case Model::MT_INFORMATION: QMessageBox::information(this, tr("Qoobar"), text);
//        break;
    case MT_WARNING: QMessageBox::warning(0,tr("Qoobar"),text);
    }
}


