/**
 * \file tagrender.cpp
 * Renders tags from placeholders.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 31 Dec 2010
 *
 * Copyright (C) 2010  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "tagrender.h"
#include "tagger.h"
#include "legendbutton.h"
#include "application.h"
#include <QFileInfo>

#include "enums.h"
#include "qoobarglobals.h"
#include "stringroutines.h"
#include "placeholders.h"
#include "model.h"

QList<QList<int> > parseNumericFunction(const QString &pattern,int length)
{

    QList<QList<int> > result;
    QRegExp startNumRx(QSL("(\\d+)(?:\\s*-\\s*)*(\\d*)"));
    QRegExp repeatRx(QSL("(\\d+)(?:\\s*:\\s*)(\\d+)"));

    int beginSec = pattern.indexOf("{",0);
    while (beginSec>=0) {
        QList<int> wholeList;

        QString wholeSection = pattern.mid(beginSec+1,
                                           pattern.indexOf(QLS("}"),beginSec)-beginSec-1);
        if (wholeSection.isEmpty()) wholeSection=QSL("1");


/**       {1-3,  1:3, 3:2; 2-3, 2:4, 3:3}
result: 1 1 1 2 3 3 2 2 2 2 3 3 3
*/
        QStringList sections=wholeSection.split(QSL(";"),QString::SkipEmptyParts);

        Q_FOREACH (const QString &sec, sections) {
            QList<int> list;
            int pos=0;
            pos = startNumRx.indexIn(sec, pos);
            if (pos!=-1) {
                int startNum=startNumRx.cap(1).toInt();
                int endNum=startNumRx.cap(2).toInt();
                if (endNum==0) endNum=length+startNum-1;

                int repeatPos=pos+startNumRx.matchedLength();
                while ((repeatPos = repeatRx.indexIn(sec, repeatPos)) != -1) {
                    int whatToRepeat=repeatRx.cap(1).toInt();
                    int repeats=repeatRx.cap(2).toInt();
                    if (repeats>1)
                        for (int i=0; i<repeats-1; ++i) list << whatToRepeat;
                    repeatPos += repeatRx.matchedLength();
                }
                for (int i=startNum; i<=endNum; ++i) list << i;
            }
            else
                for (int i=1; i<1+length; ++i) list << i;
            qSort(list);
            wholeList.append(list);
        }

        while (wholeList.size()<length) wholeList << 0;
        result << wholeList;
        beginSec = pattern.indexOf(QLS("{"),beginSec+1);
    }
    return result;
}

void renderNumericFunction(QString &fileName,int index,const QList<QList<int> > &list)
{

    int i=0;
    while (fileName.contains(QSL("{"))) {
        int beginSec = fileName.indexOf(QLS("{"),0);
        int endSec = fileName.indexOf(QLS("}"),beginSec);
        if (endSec < beginSec)
            endSec = fileName.length()-1;
        QString s=QString::number(list.at(i).value(index));
        if (s=="0") s.clear();
        fileName.replace(beginSec,endSec-beginSec+1,s);
        i++;
    }
}

/**
the function removes empty [] sections
*/
QString removeEmptySections(Tag &tag,const QList<QList<int> > &list, int index, const QString &oldPattern)
{

    QString pattern = oldPattern;

    int startPos=0;
    for (int i=0; i<list.size(); ++i) {
        startPos=pattern.indexOf(QLS("{"),startPos);
        if (startPos>=0) {
            int endPos=pattern.indexOf(QLS("}"),startPos);
            if (endPos<0) endPos = pattern.length();

            if (list.at(i).at(index)==0) {
                pattern.remove(startPos,endPos-startPos+1);
                startPos=0;
            }
            else
                startPos=endPos;
        }
    }


    while (pattern.contains(QSL("["))) {
        int begin = pattern.indexOf(QLS("["),0);
        int end = pattern.indexOf(QLS("]"),begin);
        if (end < begin)
            end = pattern.length()-1;
        QString sec = pattern.mid(begin+1,end-begin-1);

        for (int i=COMPOSER; i<TAGSCOUNT; ++i) {
            if (tag.tag(i).isEmpty())
                sec.remove(QSL("%") + Placeholders::idToPlaceholder(i));
        }

        if (tag.tag(TRACKNUMBER).isEmpty()) sec.remove(QSL("%N"));
        if (tag.bitrate().isEmpty()) sec.remove(QSL("%b"));

        if (!sec.contains(QSL("%")) && !sec.contains(QSL("{")))
            sec.clear();
        pattern.replace(begin,end-begin+1,sec);
    }
    return pattern;
}

void SimpleTagsRenderer::updateTags()
{
    int size = m->selectionSize();
    if (m_pattern.isEmpty()) {
        newTags = QVector<QString>(size, m_pattern).toList();
        return;
    }
    newTags.clear();
    preprocess();

    QList<QList<int> > res;
    if (m_pattern.contains(QSL("{")))
        res=parseNumericFunction(m_pattern, size);

    QRegExp re(QSL("%(.)"));

    for (int index=0; index<size; ++index) {
        Tag tag = m->fileAtSelection(index);
        QString currentPattern=m_pattern;
        if (m_pattern.contains(QSL("[")))
            currentPattern=removeEmptySections(tag,res,index,m_pattern);

        if (currentPattern.contains(QSL("%"))) {
            int pos = 0;
            while ((pos = re.indexIn(currentPattern, pos)) != -1) {
                QString cap=re.cap(1);
                int index=Placeholders::placeholderToId(cap);
                if (index>=0) {
                    QString tagValue=tag.nonEmptyTag(index,QSL("_"));
                    if (cap=="N" && tagValue!="_") tagValue=tagValue.rightJustified(2,QChar('0'));
                    currentPattern.replace(pos,2,tagValue);
                    pos += tagValue.length();
                }
                else pos += 2;
            }

            currentPattern.replace(QLS("%l"),formatLength(tag.length()));
            currentPattern.replace(QLS("%L"),QString::number(tag.length()));
            currentPattern.replace(QLS("%b"),tag.bitrate());
            currentPattern.replace(QLS("%f"),tag.fileName());
            currentPattern.replace(QLS("%r"),QString::number(tag.sampleRate()));
            currentPattern.replace(QLS("%h"),QString::number(tag.channels()));
        }

        if (currentPattern.contains(QSL("{")))
            renderNumericFunction(currentPattern,index,res);
        newTags << currentPattern;
    }
    postprocess();
}

void FileNameRenderer::updateTags()
{

    SimpleTagsRenderer::updateTags();
    if (m_pattern.isEmpty()) return;

    prefixes.clear();
    for (int i=0; i<newTags.size(); ++i) {
        QString s=newTags.at(i);
        if (App->replaceWinChars)
            s=replaceWinChars(s, App->winCharsReplacer, true);

        //slashes in tags
        s.replace(QLS("/"),App->winCharsReplacer);

        //some operations we perform only on file names and not paths
        QStringList list=s.split(QChar(0x263c));
        QString fileName=list.takeLast();

        if (App->replaceSpaces) {
            fileName.replace(QLS(" "),App->spacesReplacer);
            fileName.replace(QLS("\t"),App->spacesReplacer);
            if (App->applyToFolders) {
                list.replaceInStrings(QSL(" "),App->spacesReplacer);
                list.replaceInStrings(QSL("\t"),App->spacesReplacer);
            }
        }

        if (App->removeDiacritics) {
            fileName=removeDiacritics(fileName);
            if (App->applyToFolders) {
                QMutableStringListIterator i(list);
                while (i.hasNext()) {
                    i.next();
                    i.setValue(removeDiacritics(i.value()));
                }
            }
        }

        if (App->changeCase) {
            fileName=changeCase(fileName,(Case::Case)App->ccase);
            if (App->applyToFolders) {
                QMutableStringListIterator i(list);
                while (i.hasNext()) {
                    i.next();
                    i.setValue(changeCase(i.value(),(Case::Case)App->ccase));
                }
            }
        }

        QFileInfo fi(m->fileAtSelection(i).fullFileName());

        if (App->trimFileLength && App->maximumLength>0) {
            int suffixLength=fi.suffix().length()+1;
            if (App->maximumLength-suffixLength>0) fileName.truncate(App->maximumLength-suffixLength);
        }
        QString pathPrefix=App->destinationFolder;
        if (pathPrefix.isEmpty())
            pathPrefix = fi.canonicalPath();
        prefixes << pathPrefix;
        list.prepend(pathPrefix);
        list.append(QString("%1.%2").arg(fileName).arg(fi.suffix()));
        newTags[i] = list.join(QSL("/"));
    }
}

void FileNameRenderer::preprocess()
{

    SimpleTagsRenderer::preprocess();

    m_pattern.replace(QLS("\\"),QLS("/"));
    m_pattern.replace(QLS("/"),QString(QChar(0x263c)));
}

void SimpleTagsRenderer::preprocess()
{

    m_pattern.replace(QLS("\\["),QChar(0x2590));
    m_pattern.replace(QLS("\\]"),QChar(0x2591));
    m_pattern.replace(QLS("[["),QChar(0x2590));
    m_pattern.replace(QLS("]]"),QChar(0x2591));
}

void SimpleTagsRenderer::postprocess()
{
    newTags.replaceInStrings(QChar(0x2590),QSL("["));
    newTags.replaceInStrings(QChar(0x2591),QSL("]"));
}
